/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/macosx/view.m: Implementation of SILView, a subclass of
 * NSView with OpenGL support.
 */

#define IN_SYSDEP

#import "src/base.h"
#import "src/input.h"
#import "src/memory.h"
#import "src/sysdep.h"
#import "src/sysdep/darwin/time.h"
#import "src/sysdep/macosx/view.h"
#import "src/sysdep/opengl/opengl.h"

#import "src/sysdep/macosx/osx-headers.h"
#import <AppKit/NSOpenGL.h>
#import <AppKit/NSView.h>
#import <OpenGL/OpenGL.h>

/*************************************************************************/
/***************************** Public methods ****************************/
/*************************************************************************/

@implementation SILView

/*-----------------------------------------------------------------------*/

- (id)initWithFrame:(NSRect)frame pixelFormat:(NSOpenGLPixelFormat *)format
{
    self = [super initWithFrame:frame];
    if (!self) {
        return nil;
    }

    pixel_format = [format retain];

    return self;
}

/*-----------------------------------------------------------------------*/

- (void)dealloc
{
    [pixel_format release];
    [super dealloc];
}

/*-----------------------------------------------------------------------*/

- (BOOL)createGLContext:(BOOL)forRendering
{
    if (context && forRendering) {
        /* We don't currently support rendering from multiple threads. */
        return FALSE;
    }

    /* Create the OpenGL context. */
    NSOpenGLContext *new_context =
        [[NSOpenGLContext alloc] initWithFormat:pixel_format shareContext:nil];
    if (!new_context) {
        return FALSE;
    }
    [new_context makeCurrentContext];

    if (forRendering) {
        /* Save this as the primary context. */
        context = new_context;
        /* Associate the context with the view. */
        [context setView:self];
        /* Pass view frame change notifications to the context. */
        [[NSNotificationCenter defaultCenter] addObserver:context
            selector:@selector(update)
            name:NSViewGlobalFrameDidChangeNotification
            object:self];
    }

    return TRUE;
}

/*-----------------------------------------------------------------------*/

- (void)destroyGLContext
{
    NSOpenGLContext *this_context = [NSOpenGLContext currentContext];
    [NSOpenGLContext clearCurrentContext];
    if (this_context) {
        if (this_context == context) {
            [[NSNotificationCenter defaultCenter] removeObserver:context];
            context = nil;
        }
        [this_context release];
    }
}

/*-----------------------------------------------------------------------*/

@end

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
